/*
 * GQview
 * (C) 2001 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */


#include "gqview.h"

#include "collect.h"
#include "dnd.h"
#include "filelist.h"
#include "img-main.h"
#include "menu.h"
#include "preferences.h"
#include "rcfile.h"
#include "slideshow.h"
#include "utilops.h"
#include "window.h"
#include "ui_help.h"
#include "ui_fileops.h"
#include "ui_tabcomp.h"
#include "ui_utildlg.h"

#include <gdk/gdkkeysyms.h> /* for keyboard values */

#include "icons/icon.xpm"


/*
 *-----------------------------------------------------------------------------
 * misc (public)
 *-----------------------------------------------------------------------------
 */ 
typedef struct _WindowIconData WindowIconData;
struct _WindowIconData
{
	const char **icon;
	gchar *path;
};

static void window_set_icon_cb(GtkWidget *widget, gpointer data)
{
	WindowIconData *wid = data;
	GdkPixbuf *pb;
	GdkPixmap *pixmap;
	GdkBitmap *mask;

	if (wid->icon)
		{
		pb = gdk_pixbuf_new_from_xpm_data(wid->icon);
		}
	else
		{
		pb = gdk_pixbuf_new_from_file(wid->path);
		}

	g_free(wid->path);
	g_free(wid);

	if (!pb) return;

	gdk_pixbuf_render_pixmap_and_mask(pb, &pixmap, &mask, 128);
	gdk_pixbuf_unref(pb);

	gdk_window_set_icon(widget->window, NULL, pixmap, mask);
	/* apparently, gdk_window_set_icon does not ref the pixmap and mask, so don't unref it (leak?) */
}

void window_set_icon(GtkWidget *window, const char **icon, const gchar *file)
{
	WindowIconData *wid;

	if (!icon && !file) icon = (const char **)icon_xpm;

	wid = g_new0(WindowIconData, 1);
	wid->icon = icon;
	wid->path = g_strdup(file);

	gtk_signal_connect(GTK_OBJECT(window), "realize",
			   GTK_SIGNAL_FUNC(window_set_icon_cb), wid);
}

gfloat get_zoom_increment(void)
{
	return (zoom_increment ? zoom_increment / 10.0 : 1.0);
}

static gint file_extension_match(const gchar *path, const gchar *ext)
{
	gint p;
	gint e;

	if (!path) return FALSE;
	if (!ext) return TRUE;

	p = strlen(path);
	e = strlen(ext);

	return (p > e && strncasecmp(path + p - e, ext, e) == 0);
}
/*
 *-----------------------------------------------------------------------------
 * help window
 *-----------------------------------------------------------------------------
 */ 

static GtkWidget *help_window = NULL;

static void help_window_destroy_cb(GtkWidget *window, gpointer data)
{
	help_window = NULL;
}

void help_window_show(const gchar *key)
{
	if (help_window)
		{
		gdk_window_raise(help_window->window);
		if (key) help_window_set_key(help_window, key);
		return;
		}

	help_window = help_window_new(_("Help - GQview"), "GQview", "help",
                                      GQVIEW_HELPDIR "/README", key);
	gtk_signal_connect(GTK_OBJECT(help_window), "destroy", help_window_destroy_cb, NULL);
}


/*
 *-----------------------------------------------------------------------------
 * external editor start routines (public)
 *-----------------------------------------------------------------------------
 */ 

void start_editor_from_file(gint n, const gchar *path)
{
	gchar *cmd;
	if (!path) return;
	cmd = g_strdup_printf("%s \"%s\" &", editor_command[n], path);
	if (debug) printf(_("GQview running: %s\n"),cmd);
	system(cmd);
	g_free(cmd);
}

void start_editor_from_image(gint n)
{
	start_editor_from_file(n, main_image_get_path());
}

void start_editor_from_path_list(gint n, GList *list)
{
	gchar *cmd;
	gchar *buf;
	GList *work;

	if (!list) return;

	work = list;
	cmd = g_strconcat(editor_command[n], " ", NULL);
	while(work)
		{
		buf = cmd;
		cmd = g_strconcat(buf, "\"", work->data, "\" ", NULL);
		g_free(buf);
		work = work->next;
		}
	buf = cmd;
	cmd = g_strconcat(buf, "&", NULL);
	g_free(buf);
	if (debug) printf(_("GQview running: %s\n"), cmd);
	system(cmd);
	g_free(cmd);
}

void start_editor_from_list(gint n)
{
	GList *list;

	list = file_get_selected_list();
	if (!list) return;
	start_editor_from_path_list(n, list);
	path_list_free(list);
}

/*
 *-----------------------------------------------------------------------------
 * keyboard functions
 *-----------------------------------------------------------------------------
 */

void keyboard_scroll_calc(gint *x, gint *y, GdkEventKey *event)
{
	static gint delta = 0;
	static guint32 time_old = 0;
	static guint keyval_old = 0;

	if (progressive_key_scrolling)
		{
		guint32 time_diff;

		time_diff = event->time - time_old;

		/* key pressed within 125ms ? (1/8 second) */
		if (time_diff > 125 || event->keyval != keyval_old) delta = 0;

		time_old = event->time;
		keyval_old = event->keyval;

		delta += 2;
		}
	else
		{
		delta = 8;
		}

	*x = *x * delta;
	*y = *y * delta;
}

gint key_press_cb(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
	gint stop_signal = FALSE;
	gint x = 0;
	gint y = 0;

	if (GTK_WIDGET_HAS_FOCUS(path_entry))
		{
		if (event->keyval == GDK_Escape)
			gtk_entry_set_text(GTK_ENTRY(path_entry), current_path);
		return stop_signal;
		}

	if (full_screen_window || GTK_WIDGET_HAS_FOCUS(main_image->widget))
		{
		switch (event->keyval)
			{
			case GDK_Left: case GDK_KP_Left:
				x -= 1;
				stop_signal = TRUE;
				break;
			case GDK_Right: case GDK_KP_Right:
				x += 1;
				stop_signal = TRUE;
				break;
			case GDK_Up: case GDK_KP_Up:
				y -= 1;
				stop_signal = TRUE;
				break;
			case GDK_Down: case GDK_KP_Down:
				y += 1;
				stop_signal = TRUE;
				break;
			case GDK_BackSpace:
			case 'B': case 'b':
				main_image_prev();
				stop_signal = TRUE;
				break;
			case GDK_space:
				main_image_next();
				stop_signal = TRUE;
				break;
			}
		}

	if ( !(event->state & GDK_CONTROL_MASK) )
	    switch (event->keyval)
		{
		case '+': case GDK_KP_Add:
			main_image_adjust_zoom(get_zoom_increment());
			stop_signal = TRUE;
			break;
		case GDK_KP_Subtract:
			main_image_adjust_zoom(-get_zoom_increment());
			stop_signal = TRUE;
			break;
		case GDK_KP_Multiply:
			main_image_set_zoom(0.0);
			stop_signal = TRUE;
			break;
		case GDK_KP_Divide: case '1':
			main_image_set_zoom(1.0);
			stop_signal = TRUE;
			break;
		case '2':
			main_image_set_zoom(2.0);
			stop_signal = TRUE;
			break;
		case '3':
			main_image_set_zoom(3.0);
			stop_signal = TRUE;
			break;
		case '4':
			main_image_set_zoom(4.0);
			stop_signal = TRUE;
			break;
		case GDK_Page_Up: case GDK_KP_Page_Up:
			main_image_prev();
			stop_signal = TRUE;
			break;
		case GDK_Page_Down: case GDK_KP_Page_Down:
			main_image_next();
			stop_signal = TRUE;
			break;
		case GDK_Home: case GDK_KP_Home:
			main_image_first();
			stop_signal = TRUE;
			break;
		case GDK_End: case GDK_KP_End:
			main_image_last();
			stop_signal = TRUE;
			break;
		case GDK_Delete: case GDK_KP_Delete:
			if (enable_delete_key)
				{
				file_util_delete(NULL, file_get_selected_list());
				stop_signal = TRUE;
				}
			break;
		case GDK_Escape:
			interrupt_thumbs();
			stop_signal = TRUE;
			break;
		case 'P': case 'p':
			main_image_slideshow_pause_toggle();
			break;
		case 'Q': case 'q':
			if (event->state == 0 || (event->state & GDK_MODIFIER_MASK) == GDK_LOCK_MASK)
				{
				exit_gqview();
				return FALSE;
				}
			break;
		}

	if (event->state & GDK_SHIFT_MASK)
		{
		x *= 3;
		y *= 3;
		}

	if (x != 0 || y!= 0)
		{
		keyboard_scroll_calc(&x, &y, event);
		main_image_scroll(x, y);
		}

	if (stop_signal) gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");

	return stop_signal;
}

/*
 *-----------------------------------------------------------------------------
 * command line parser (private) hehe, who needs popt anyway?
 *-----------------------------------------------------------------------------
 */ 

static gint startup_full_screen = FALSE;
static gint startup_in_slideshow = FALSE;
static gint startup_command_line_collection = FALSE;


static void parse_command_line_add_file(const gchar *new_path, gchar **path, gchar **file,
				        GList **list, GList **collection_list)
{
	if (file_extension_match(new_path, ".gqv"))
		{
		*collection_list = g_list_append(*collection_list, g_strdup(new_path));
		}
	else
		{
		if (!*path) *path = remove_level_from_path(new_path);
		if (!*file) *file = g_strdup(new_path);
		*list = g_list_append(*list, g_strdup(new_path));
		}
}

static void parse_command_line(int argc, char *argv[], gchar **path, gchar **file,
			       GList **cmd_list, GList **collection_list)
{
	GList *list = NULL;

	if (argc > 1)
		{
		gint i;
		gchar *base_dir = get_current_dir();
		i = 1;
		while (i < argc)
			{
			const gchar *cmd_line = argv[i];
			gchar *cmd_all = g_strconcat(base_dir, "/", cmd_line, NULL);

			if (!*path && cmd_line[0] == '/' && isdir(cmd_line))
				{
				*path = g_strdup(cmd_line);
				}
			else if (!*path && isdir(cmd_all))
				{
				*path = g_strdup(cmd_all);
				}
			else if (cmd_line[0] == '/' && isfile(cmd_line))
				{
				parse_command_line_add_file(cmd_line, path, file, &list, collection_list);
				}
			else if (isfile(cmd_all))
				{
				parse_command_line_add_file(cmd_all, path, file, &list, collection_list);
				}
			else if (strcmp(cmd_line, "--debug") == 0)
				{
				/* we now increment the debug state for verbosity */
				debug++;
				printf("debugging output enabled (level %d)\n", debug);
				}
			else if (strcmp(cmd_line, "+t") == 0 ||
				 strcmp(cmd_line, "--with-tools") == 0)
				{
				tools_float = FALSE;
				tools_hidden = FALSE;
				}
			else if (strcmp(cmd_line, "-t") == 0 ||
				 strcmp(cmd_line, "--without-tools") == 0)
				{
				tools_hidden = TRUE;
				}
			else if (strcmp(cmd_line, "-f") == 0 ||
				 strcmp(cmd_line, "--fullscreen") == 0)
				{
				startup_full_screen = TRUE;
				}
			else if (strcmp(cmd_line, "-s") == 0 ||
				 strcmp(cmd_line, "--slideshow") == 0)
				{
				startup_in_slideshow = TRUE;
				}
			else if (strcmp(cmd_line, "-l") == 0 ||
				 strcmp(cmd_line, "--list") == 0)
				{
				startup_command_line_collection = TRUE;
				}
			else if (strcmp(cmd_line, "-v") == 0 ||
				 strcmp(cmd_line, "--version") == 0)
				{
				printf("GQview %s\n", VERSION);
				exit (0);
				}
			else if (strcmp(cmd_line, "-h") == 0 ||
				 strcmp(cmd_line, "--help") == 0)
				{
				printf("GQview %s\n", VERSION);
				printf(_("Usage: gqview [options] [path]\n\n"));
				printf(_("valid options are:\n"));
				printf(_("  +t, --with-tools           force show of tools\n"));
				printf(_("  -t, --without-tools        force hide of tools\n"));
				printf(_("  -f, --fullscreen           start in full screen mode\n"));
				printf(_("  -s, --slideshow            start in slideshow mode\n"));
				printf(_("  -l, --list                 open collection window for command line\n"));
				printf(_("  --debug                    turn on debug output\n"));
				printf(_("  -v, --version              print version info\n"));
				printf(_("  -h, --help                 show this message\n\n"));
				exit (0);
				}
			else 
				{
				printf(_("invalid or ignored: %s\nUse --help for options\n"), cmd_line);
				}
			g_free(cmd_all);
			i++;
			}
		g_free(base_dir);
		parse_out_relatives(*path);
		parse_out_relatives(*file);
		}

	if (list && list->next)
		{
		GList *work;

		work = list;
		while (work)
			{
			parse_out_relatives((gchar *)work->data);
			work = work->next;
			}

		*cmd_list = list;
		}
	else
		{
		path_list_free(list);
		*cmd_list = NULL;
		}
}

/*
 *-----------------------------------------------------------------------------
 * startup, init, and exit
 *-----------------------------------------------------------------------------
 */ 

#define RC_HISTORY_NAME "history"

static void keys_load(void)
{
	gchar *path;

	path = g_strconcat(homedir(), "/", GQVIEW_RC_DIR, "/", RC_HISTORY_NAME, NULL);
	history_list_load(path);
	g_free(path);
}

static void keys_save(void)
{
	gchar *path;

	path = g_strconcat(homedir(), "/", GQVIEW_RC_DIR, "/", RC_HISTORY_NAME, NULL);
	history_list_save(path);
	g_free(path);
}

static void check_for_home_path(gchar *path)
{
	gchar *buf;
	buf = g_strconcat(homedir(), "/", path, NULL);
	if (!isdir(buf))
		{
		printf(_("Creating GQview dir:%s\n"), buf);
		if (mkdir (buf, 0755) < 0)
			{
			printf(_("Could not create dir:%s\n"), buf);
			}
		}
	g_free(buf);
}

static void setup_default_options(void)
{
	gint i;

	for(i=0; i<8; i++)
		{
		editor_name[i] = NULL;
		editor_command[i] = NULL;
		}

	editor_name[0] = g_strdup(_("The Gimp"));
	editor_command[0] = g_strdup("gimp-remote -n");

	editor_name[1] = g_strdup(_("Electric Eyes"));
	editor_command[1] = g_strdup("ee");

	editor_name[2] = g_strdup(_("XV"));
	editor_command[2] = g_strdup("xv");

	editor_name[3] = g_strdup(_("Xpaint"));
	editor_command[3] = g_strdup("xpaint");

	custom_filter = g_strdup(".eim;");
}

static void exit_gqview_final(void)
{
	gchar *path;

	gdk_window_get_root_origin (mainwindow->window, &main_window_x, &main_window_y);
	gdk_window_get_size(mainwindow->window, &main_window_w, &main_window_h);

	if (!toolwindow || (toolwindow && GTK_WIDGET_VISIBLE(toolwindow)) )
		{
		main_window_vdivider_pos = dir_clist->allocation.height;
		}

	if (toolwindow && GTK_WIDGET_VISIBLE(toolwindow))
		{
		gdk_window_get_root_origin (toolwindow->window, &float_window_x, &float_window_y);
		gdk_window_get_size(toolwindow->window, &float_window_w, &float_window_h);
		}
	save_options();
	keys_save();

	path = g_strconcat(homedir(), "/", GQVIEW_RC_DIR, "/accels", NULL);
	gtk_item_factory_dump_rc(path, NULL, FALSE);
	g_free(path);

	gtk_main_quit();
}

static GenericDialog *exit_dialog = NULL;

static void exit_confirm_cancel_cb(GenericDialog *gd, gpointer data)
{
	exit_dialog = NULL;
	generic_dialog_close(gd);
}

static void exit_confirm_exit_cb(GenericDialog *gd, gpointer data)
{
	exit_dialog = NULL;
	generic_dialog_close(gd);
	exit_gqview_final();
}

static gint exit_confirm_dlg(void)
{
	if (exit_dialog)
		{
		gdk_window_raise(exit_dialog->dialog->window);
		return TRUE;
		}

	if (!collection_window_modified_exists()) return FALSE;

	exit_dialog = generic_dialog_new(_("GQview - exit"), _("Collections have been modified.\nExit anyway?"),
				"GQview", "exit", FALSE,
				exit_confirm_cancel_cb, NULL);
	generic_dialog_add(exit_dialog, _("Exit"), exit_confirm_exit_cb, TRUE);

	gtk_widget_show(exit_dialog->dialog);

	return TRUE;
}

void exit_gqview(void)
{
	main_image_full_screen_stop();

	if (exit_confirm_dlg()) return;

	exit_gqview_final();
}

int main (int argc, char *argv[])
{
	gchar *cmd_path = NULL;
	gchar *cmd_file = NULL;
	GList *cmd_list = NULL;
	GList *collection_list = NULL;
	gchar *buf;

	/* setup locale, i18n */
	gtk_set_locale();
	bindtextdomain (PACKAGE, LOCALEDIR);
	textdomain (PACKAGE);

	/* setup random seed for random slideshow */
        srand(time(NULL));

	check_for_home_path(GQVIEW_RC_DIR);
	check_for_home_path(GQVIEW_RC_DIR_THUMBS);
	check_for_home_path(GQVIEW_RC_DIR_COLLECTIONS);

	/* backwards compatibility */
	buf = g_strconcat(homedir(), "/.gqviewrc", NULL);
	if (isfile(buf))
		{
		gchar *new_buf;
		new_buf = g_strconcat(homedir(), "/", GQVIEW_RC_DIR, "/", RC_FILE_NAME, NULL);
		printf("moving old location of .gqviewrc to new location:\n%s\n", new_buf);
		if (rename(buf, new_buf) != 0)
			{
			printf("Warning: move of .gqviewrc failed!\n");
			}
		g_free(new_buf);
		}
	g_free(buf);

	buf = g_strconcat(homedir(), "/.gqview_thmb", NULL);
	if (isdir(buf))
		{
		gchar *new_buf;
		new_buf = g_strconcat(homedir(), "/", GQVIEW_RC_DIR_THUMBS, NULL);
		printf("moving old location of .gqview_thmb to new location:\n%s\n", new_buf);
		if (rename(buf, new_buf) != 0)
			{
			printf("Warning: move of .gqview_thmb failed!\n");
			}
		g_free(new_buf);
		}
	g_free(buf);

	gtk_init (&argc, &argv);

	gdk_rgb_init();


	/* push the correct color depths to gtk, (for 8-bit psuedo color displays)
	 * they should be popped, too, I guess...
	 */
	gtk_widget_push_visual(gdk_rgb_get_visual());
	gtk_widget_push_colormap(gdk_rgb_get_cmap());

	setup_default_options();
	load_options();
	keys_load();

	buf = g_strconcat(homedir(), "/", GQVIEW_RC_DIR, "/accels", NULL);
	gtk_item_factory_parse_rc(buf);
	g_free(buf);

	parse_command_line(argc, argv, &cmd_path, &cmd_file, &cmd_list, &collection_list);

	if (cmd_path)
		current_path = g_strdup(cmd_path);
	else if (startup_path_enable && startup_path && isdir(startup_path))
		current_path = g_strdup(startup_path);
	else
		current_path = get_current_dir();

	create_main_window();
	update_edit_menus(mainwindow_accel_grp);
	rebuild_file_filter();
	filelist_set_sort_method(file_sort_method, FALSE);
	filelist_refresh();

	init_dnd();

	if (cmd_list)
		{
		CollectionData *cd;
		GList *work;

		if (startup_command_line_collection)
			{
			CollectWindow *cw;

			cw = collection_window_new("");
			cd = cw->cd;
			}
		else
			{
			cd = collection_new("");	/* if we pass NULL, untitled counter is falsely increm. */
			}

		g_free(cd->path);
		cd->path = NULL;
		g_free(cd->name);
		cd->name = g_strdup(_("Command line"));

		collection_path_changed(cd);

		work = cmd_list;
		while (work)
			{
			collection_add(cd, (gchar *)work->data, FALSE);
			work = work->next;
			}

		if (cd->list) main_image_change_to_from_collection(cd, cd->list->data);
#if 0
		/* mem leak, we never unref this collection when !startup_command_line_collection
		 * (the image view of the main window does not hold a ref to the collection)
		 * this is sort of unavoidable, for if it did hold a ref, next/back
		 * may not work as expected when closing collection windows.
		 */
		collection_unref(cd);
#endif
		}
	else if (cmd_file)
		{
		filelist_change_to_full_path(cmd_file);
		}
	else
		{
		main_image_change_to(NULL, NULL);
		}

	if (collection_list)
		{
		GList *work;

		work = collection_list;
		while (work)
			{
			const gchar *path = work->data;
			work = work->next;

			collection_window_new(path);
			}
		}

	g_free(cmd_path);
	g_free(cmd_file);
	path_list_free(cmd_list);
	path_list_free(collection_list);

	if (startup_full_screen) main_image_full_screen_toggle();
	if (startup_in_slideshow) main_image_slideshow_start();

	gtk_main ();
	return 0;
}


